import scanpy as sc
import numpy as np
import pandas as pd
import os
import pickle
from scipy.stats import rankdata
import scipy.sparse as sp
from scipy.sparse import csr_matrix
import anndata
import os
from ..dynamic_graphs.distDistance import getClusterRepresentation

def get_data_file_path(filename):
    '''
    get the path of data file

    parameters
    ------------------
    filename: str
        name of the file

    return
    ------------------
    file_path: str
        path of the file
    '''
    file_path = os.path.join(os.path.dirname(__file__), 'data', filename)
    return file_path

def split_dataset_into_stage(adata_path, folder, key):
    '''
    split dataset into stages and write to the path
    
    parameters
    ------------------
    adata: IPF database
    key: key of the dataset
    path: path to write the dataset

    '''
    adata = sc.read_h5ad(adata_path)
    for each in list(adata.obs[key].unqiue()):
        adata_temp = adata[adata.obs[key] == each]
        adata_temp.write_h5ad(os.path.join(folder,'%s.h5ad'%each),compression='gzip',compression_opts=9)
def transfer_to_ranking_score(gw):
    '''
    transfer gene weight to ranking score

    parameters
    ------------------
    gw: np.array
        gene weight of each gene in each cell

    return
    ------------------
    score: np.array
        ranking score of each gene in each cell
    '''

    od = gw.shape[1]-rankdata(gw,axis=1)+1
    score = 1+1/np.power(od,0.5)
    
    return score

def clustertype40(adata):
    '''
    annotate the cluster with cells >40% if no one >40%, annotate with the highest one
    
    parameters
    ------------------
    adata: anndata of one cluster
    
    return
    ---------------- 
    anootate:
        The cluster type
    '''
    dic = {}
    total = 0
    
    for each in adata.obs['name.simple']:
        if each not in dic.keys():
            dic[each]=1
        else:
            dic[each]+=1
        total+=1
    #print(dic)
   
    anootate = ''
    
    flag = False #flag to see if there are more than 1 cell types > 40%
    for each in list(dic.keys()):
        
        if dic[each] > total*0.5:
            if flag == False:
                anootate+=each
                flag = True
            else:
                anootate+='/'+each
    if flag == False:
        for each in list(dic.keys()):
        
            if dic[each] > total*0.4:
                if flag == False:
                    anootate+=each
                    flag = True
                else:
                    anootate+='/'+each
        if flag == False:
            for each in list(dic.keys()):
        
                if dic[each] > total*0.3:
                    if flag == False:
                        anootate+=each
                        flag = True
                    else:
                        anootate+='/'+each
        if flag == False:
            for each in list(dic.keys()):
        
                if dic[each] > total*0.2:
                    if flag == False:
                        anootate+=each
                        flag = True
                    else:
                        anootate+='/'+each
        if flag == False:
            anootate = 'Mixed'
    
    return anootate

def changeCluster(adata, ids, newIds):
    '''
    re-assign cluster id to cells

    parameters
    ------------------
    adata: anndata
        IPF database
    ids: int
        cell id in one stage needed to be changed 
    newIds: int
        new cluster id of cells

    return
    ------------------
    adata: anndata
        the updated IPF data
    '''

    for i, cluster in enumerate(ids):
        for j, each in enumerate(cluster):
            #print('vamos[i][j]',vamos[i][j])
            adata.obs['leiden'][each] = newIds[i][j]
    return adata
def extracth5adcluster(data, ID):
    '''
    extract data from a certain cluster

    parameters
    ------------------
    data: anndata
        data (H5AD class) of a certain stage
    ID: int
        the id of cluster
    
    return
    ------------------
    splitData: anndata
        data of a certain cluster
    splitDataID: list
        index of cells in a certain cluster
    '''
    splitDataID = []
    splitData = []
    for i in range(len(data)):
        
        if int(data.obs['leiden'][i]) == ID:
            splitDataID.append(i)
    splitData = data[splitDataID]
    
    return splitData,splitDataID

def extracth5adclusterids(data, ID):
    '''
    extract index from a certain cluster
    parameters
    ------------------
    data: anndata
        data (H5AD class) of a certain stage
    ID: int
        the id of cluster
    
    return:
    splitDataID: list
        index of cells in a certain cluster
    '''
    splitDataID = []
    splitData = []
    for i in range(len(data)):
        
        if int(data.obs['leiden'][i]) == ID:
            splitDataID+=data[i].obs.index.tolist()
    splitData = data[splitDataID]
    
    return splitDataID

def updateAttributes(adata,reps):
    '''
    update stage, cluster id of each stage, top 100 differential genes, cell types of clusters to anndata

    parameters
    ------------------
    adata: anndata
        anndata of database
    reps: list
        representations of sampled gaussian data points
    
    
    return
    ------------------ 
    adata: anndata
        updated anndata of database
    rep: list
        representations of sampled gaussian data points
    average_cluster: np.array
        average expression of each cluster
    '''
    print('top gene')
    sc.tl.rank_genes_groups(adata, 'leiden', method='wilcoxon',n_genes=100)
    print('done')
    clusters = set(adata.obs['leiden'])
    average_cluster= []
    TDG = []
    rep = []
    celltypes = []
    
    # adata.obs['name.simple'] = adata.obs['name.simple'].astype('string')
    for i in range(len(clusters)):
        
        clusteradata,ids = extracth5adcluster(adata,i)
        clutertype = clustertype40(clusteradata) #use >40 strategy
        celltypes.append(clutertype)
        one_mean = np.mean(clusteradata.X,axis =0)
        one_mean = np.reshape(np.array(one_mean),-1)
        average_cluster.append(one_mean)
        adata.obs['ident'] = adata.obs['ident'].astype(str)
        TDG.append(list(adata.uns['rank_genes_groups']['names'][str(i)]))

        rep.append(getClusterRepresentation(reps[0][ids],reps[1][ids]))
        ids = extracth5adclusterids(adata,i)
        adata.obs.loc[adata.obs.index.isin(ids), "ident"] = clutertype
        # print('cluster type showing')
        # print(clutertype)
        # print(adata[ids].obs['ident'])
    adata.obs['ident'] = adata.obs['ident'].astype(str)
    adata.uns['clusterType'] = celltypes
    adata.uns['topGene'] = TDG
    return adata,average_cluster,rep

def saveRep(rep,midpath,iteration):
    '''
    write latent representations 'Z' of all stages to disk in iterative training

    parameters
    ------------------
    rep: list
        representations of sampled gaussian data points
    midpath: str
        directory to the task
    iteration: int
        iteration number
    '''
    np.save(os.path.join(midpath,str(iteration)+'/stagedata/rep.npy'),np.array(rep,dtype=object))

def get_all_adj_adata(adatas):
    '''
    merge all stages to a whole one

    parameters
    ------------------
    adatas: list
        list of anndata of each stage

    return
    ------------------
    adata: anndata
        anndata of the whole dataset
    '''
    data_size = []
    obs = []
    for i, each in enumerate(adatas):
        each.obs['stage'] = i
        obs.append(each.obs)
        each.obsp['gcn_connectivities'] = each.obsp['gcn_connectivities'].tocoo()
        if i == 0:
            if sp.isspmatrix(each.X):
                X = each.X
            else:
                X = csr_matrix(each.X)
            data_size.append(each.obsp['gcn_connectivities'].shape[0])
            gcn_data = each.obsp['gcn_connectivities'].data.tolist()
            col = each.obsp['gcn_connectivities'].col.tolist()
            row = each.obsp['gcn_connectivities'].row.tolist()
            if 'geneWeight' in each.layers.keys():
                geneWeight = each.layers['geneWeight']
            if 'X_pca' in each.obsm.keys():
                pca = each.obsm['X_pca']

        else:
            X = sp.vstack((X, each.X), format='csr')
            
            gcn_data += each.obsp['gcn_connectivities'].data.tolist()
            col += (data_size[i-1] + each.obsp['gcn_connectivities'].col).tolist()
            row += (data_size[i-1] + each.obsp['gcn_connectivities'].row).tolist()
            data_size.append(each.obsp['gcn_connectivities'].shape[0] + data_size[i - 1])
            if 'geneWeight' in each.layers.keys():
                geneWeight = sp.vstack((geneWeight, each.layers['geneWeight']), format='csr')
            if 'X_pca' in each.obsm.keys():
                pca = np.concatenate((pca,each.obsm['X_pca']),axis=0)
        each.obsp['gcn_connectivities'] = each.obsp['gcn_connectivities'].tocsr()
    
    obs = pd.concat(obs)
    variable = adatas[0].var 
    adata = anndata.AnnData(X=X,obs=obs,var=variable)
    gcn = csr_matrix((gcn_data, (row,col)), shape=(adata.X.shape[0],adata.X.shape[0]))

    if 'geneWeight' in each.layers.keys():
        adata.layers['geneWeight'] = geneWeight
    adata.obsp['gcn_connectivities'] = gcn
    if 'X_pca' in each.obsm.keys():
        adata.obsm['X_pca'] = pca
  
    return adata

def mergeAdata(path,total_stages):
    '''
    merge all stages to a whole one dataset and write to disk

    parameters
    ------------------
    path: str
        directory to the task
    total_stages: int
        total number of stages
    '''
    #read stage datasets
    adatas = []
    for i in range(total_stages):
        adata = sc.read_h5ad(os.path.join(path, 'stagedata/%d.h5ad'%i))
        adata.obs['stage'] = i
        adatas.append(adata)
    for i, each in enumerate(adatas):
        adatas[i].obsp['gcn_connectivities'] = adatas[i].obsp['gcn_connectivities'].tocoo()

    
    gcn_data = adatas[0].obsp['gcn_connectivities'].data.tolist()
    col = adatas[0].obsp['gcn_connectivities'].col.tolist()
    row = adatas[0].obsp['gcn_connectivities'].row.tolist()
    for i in range(1,total_stages):

        current_shape = sum([adatas[i].obsp['gcn_connectivities'].shape[0] for i in range(i)])
        col+=(current_shape+ adatas[i].obsp['gcn_connectivities'].col).tolist()
        row+=(current_shape+ adatas[i].obsp['gcn_connectivities'].row).tolist()
        gcn_data+=adatas[i].obsp['gcn_connectivities'].data.tolist()

    #get X
    if sp.isspmatrix(adatas[0].X):
        X = adatas[0].X
    else:
        X = csr_matrix(adatas[0].X)
    for i in range(1,total_stages):
        X = sp.vstack((X, adatas[i].X), format='csr')
    
    #get geneWeight
    geneWeight = adatas[0].layers['geneWeight']
    for i in range(1,total_stages):
        geneWeight = sp.vstack((geneWeight, adatas[i].layers['geneWeight']), format='csr')

    #get var
    if 'concat' in adatas[0].layers.keys():
        concat = adatas[0].layers['concat']
        for i in range(1,total_stages):
            concat = sp.vstack((concat, adatas[i].layers['concat']), format='csr')

    
    variable = adatas[0].var 
    
    #get obs
    obs = [each.obs for each in adatas]
    
    obs = pd.concat(obs)

    #get adata.ump['z']
    Z = adatas[0].obsm['z']
    for i in range(1,total_stages):
        Z = np.concatenate((Z,adatas[i].obsm['z']),axis=0)
    #get adata.ump['umap']
    umap = adatas[0].obsm['X_umap']
    for i in range(1,total_stages):
        umap = np.concatenate((umap,adatas[i].obsm['X_umap']),axis=0)
    #get top Gene
    topGene = {}
    top_gene_fold_change = {}
    top_gene_pvals_adj = {}
    clustertype = {}
    for i in range(total_stages):
        topGene[str(i)] = adatas[i].uns['topGene']
        clustertype[str(i)]=adatas[i].uns['clusterType']
    #top gene fold_change 
        adatas[i].uns['logfoldchanges'] = []
        adatas[i].uns['top_gene_pvals_adj'] = []
        for j in set(adatas[i].obs['leiden']):
            adatas[i].uns['logfoldchanges'].append(adatas[i].uns['rank_genes_groups']['logfoldchanges'][str(j)])
            adatas[i].uns['top_gene_pvals_adj'].append(adatas[i].uns['rank_genes_groups']['pvals_adj'][str(j)])

    #get uns.edges
    edges = eval(open(os.path.join(path,'edges.txt')).read())
    clusterType = clustertype
    #build new anndata and assign attribtues and write dataset
    adata = anndata.AnnData(X=X,obs=obs,var=variable)
    adata.layers['geneWeight'] = csr_matrix(geneWeight)
    adata.uns['clusterType']=clusterType
    adata.uns['edges']=edges
    adata.uns['topGene']=topGene
    adata.uns['top_gene_fold_change']=top_gene_fold_change
    adata.uns['top_gene_pvals_adj']=top_gene_pvals_adj
    adata.obsm['z']=Z
    adata.obsm['umap']=umap
    adata.obsm['X_umap'] = umap
    adata.obs['leiden'] = adata.obs['leiden'].astype(str)
    adata.obsp = {}

    gcn = csr_matrix((gcn_data, (row,col)), shape=(adata.X.shape[0],adata.X.shape[0]))
    adata.obsp['gcn_connectivities'] = gcn 
    attribute = adata.uns
    with open(os.path.join(path,'stagedata/org_attribute.pkl'),'wb') as f:
        pickle.dump(attribute, f)
    del adata.uns
    adata.write_h5ad(os.path.join(path,'stagedata/org_dataset.h5ad'),compression='gzip',compression_opts=9)



import scanpy as sc
import os
import gc
import torch
from torch.utils.data import DataLoader
from ..utils.gcn_utils import setup_graph
import pyro
import numpy as np
from ..utils.trainer_utils import transfer_to_ranking_score
import scipy.sparse as sp
from ..utils.h5adReader import H5ADataSet
#import variable
from ..train.customized_elbo import *
from pyro.optim import Adam
class UNAGI_trainer():
    '''
    The trainer class is the class to train the UNAGI model. The trainer class will train the model for a given number of epochs. 
    
    parameters
    ----------------
    model: object
        the model to be trained.
    modelName: str
        the name of the model.
    batch_size: int
        the batch size for training.
    epoch_initial: int
        the initial epoch for training.
    epoch_iter: int
        the total number of epochs for training.
    device: str
        the device to train the model.
    lr: float
        the learning rate for the variational autoencoder (VAE) model.
    lr_dis: float
        the learning rate for the discriminator.
    cuda: bool
        whether to use GPU for the model training. Default is True.
    

    '''
    def __init__(self,model, modelName,batch_size,epoch_initial,epoch_iter,device,lr, lr_dis,cuda=True):
        super(UNAGI_trainer, self).__init__()
        self.model = model
        self.modelName = modelName
        self.epoch_initial = epoch_initial
        self.epoch_iter = epoch_iter
        self.batch_size = batch_size
        self.cuda = cuda
        self.device = device
        self.lr = lr
        self.lr_dis = lr_dis
    def train_model(self,adata, vae, train_loader,adj, geneWeights=None, use_cuda=True):
        '''
        The function to train the model.

        parameters:
        ----------------

        adata: anndata
            the single-cell data.
        vae: model
            the model to be trained.
        train_loader: DataLoader
            the data loader for the model training.
        adj: scipy.sparse.csr_matrix
            the adjacency matrix of cell graphs.
        geneWeights: np.array
            the gene weights for the model training. Default is None.
        use_cuda: bool
            whether to use GPU for the model training. Default is True.

        return:
        ----------------
        total_epoch_loss_train: float
            the total loss for the model training.
        '''
        # initialize loss accumulator
        epoch_loss = 0.
        
        if use_cuda:
            adj=adj.to(self.device)
        
        placeholder = torch.zeros(adata.X.shape,dtype=torch.float32)
        optimizer_vae = Adam({"lr": self.lr})
            
        optimizer_dis = Adam({"lr": self.lr_dis}) 

    # do a training epoch over each mini-batch x
        epoch_loss = 0.
        loss1 = 0
        loss2 = 0
        for i, [x,neighbourhoods,idx] in enumerate(train_loader): 
            temp_x = placeholder.clone()
            start = i*self.batch_size
            if (1+i)*self.batch_size > len(adj):
                end =  len(adj)
            else:
                end = (1+i)*self.batch_size
            # neighbourhood = neighbourhoods[i]
            neighbourhood = [item for sublist in neighbourhoods for item in sublist]
            # neighbourhood = list(set([item for sublist in neighbourhoods for item in sublist]))
            temp_x[neighbourhood] = torch.Tensor(adata.X)[neighbourhood]
            x = temp_x

        # if on GPU put mini-batch into CUDA memory
            if self.cuda:
                x = x.to(self.device)
            if geneWeights is not None:
                geneWeights1 = torch.tensor(transfer_to_ranking_score(geneWeights[idx].toarray()))
                geneWeights1 = geneWeights1.to(self.device)
             
                loss = graphUpdater(myELBO(geneWeights1),vae.model,vae.guide,vae.discriminator,optimizer_vae,x,adj,i, start,end,device=self.device, second_optimizer=optimizer_dis)
        
            else:
                loss = graphUpdater(myELBO(),vae.model,vae.guide,vae.discriminator,optimizer_vae,x,adj,i, start,end,device=self.device,second_optimizer=optimizer_dis)
            
            # loss = graphUpdater(myELBO(),vae.model,vae.guide,vae.discriminator,optimizer_vae,x,adj,i, start,end,self.device=self.device,second_optimizer=optimizer_dis)
            epoch_loss += loss
            

        normalizer_train = len(train_loader.dataset)
        total_epoch_loss_train = epoch_loss / normalizer_train
        # print('loss1', loss1/normalizer_train)
        # print('loss2', loss2/normalizer_train)
        print('loss', total_epoch_loss_train)
        return total_epoch_loss_train
    def get_latent_representation(self,adata,iteration,target_dir):
        '''
        Retrieve the latent representation of the single-cell data.

        parameters:
        ----------------
        adata: anndata
            the single-cell data.   
        iteration: int
            the iteration used for analysis.
        target_dir: str
            the directory of the task.

        return:
        ----------------
        z_locs: np.array
            the mean of the latent representation.
        z_scales: np.array
            the variance of the latent representation.
        TZ: np.array
            the latent representation.
        '''

        if 'X_pca' not in adata.obsm.keys():
            sc.pp.pca(adata)
    
        if 'gcn_connectivities' in adata.obsp.keys():
            adj = adata.obsp['gcn_connectivities']
            adj = adj.asformat('coo')
        cell = H5ADataSet(adata)
        num_genes=cell.num_genes()
        placeholder = torch.zeros(adata.X.shape,dtype=torch.float32)
        cell_loader=DataLoader(cell,batch_size=self.batch_size,num_workers=0,shuffle=False)
        self.model.load_state_dict(torch.load(os.path.join(target_dir,'model_save/'+self.modelName+'_'+str(iteration)+'.pth'),map_location=self.device))
        TZ=[]
        z_locs = []
        z_scales = []
        adj = setup_graph(adj)
        adj = adj.to(self.device)
        if sp.isspmatrix(adata.X):
            adata.X = adata.X.toarray()
        for i, [x,neighbourhoods,idx] in enumerate(cell_loader):
            temp_x = placeholder.clone()
            start = i*self.batch_size
            if (1+i)*self.batch_size > len(adj):
                end =  len(adj)
            else:
                end = (1+i)*self.batch_size
            neighbourhood = [item for sublist in neighbourhoods for item in sublist]
            temp_x[neighbourhood] = torch.Tensor(adata.X)[neighbourhood]
            x = temp_x
            if self.cuda:
                x = x.to(self.device)
            _,mu, logvar,_,_  = self.model.getZ(x.view(-1, num_genes),adj,i,start, end,test=False)
            # mu, logvar = self.model.encoder(x.view(-1, num_genes),adj,idx)
            z = mu+logvar
            z_locs+=mu.detach().cpu().numpy().tolist()
            z_scales+=logvar.detach().cpu().numpy().tolist()
            TZ+=z.detach().cpu().numpy().tolist()
        z_locs = np.array(z_locs)
        z_scales = np.array(z_scales)
        TZ = np.array(TZ)
        return z_locs, z_scales, TZ

    def train(self, adata, iteration, target_dir, is_iterative=False):
        '''
        The function to train the model.

        parameters
        ----------------
        adata: anndata
            the single-cell data.
        iteration: int
            the iteration used for analysis.
        target_dir: str 
            the directory of the task.
        is_iterative: bool
            whether to use the iterative training strategy. Default is False.
        '''
        
        assert 'X_pca' in adata.obsm.keys(), 'PCA is not performed'
        if 'X_pca' not in adata.obsm.keys():
            sc.tl.pca(adata, svd_solver='arpack')

        if 'gcn_connectivities' in adata.obsp.keys():
            adj = adata.obsp['gcn_connectivities']
            adj = adj.asformat('coo')
        adj = setup_graph(adj)
        if is_iterative:
            geneWeights = adata.layers['geneWeight']
            cell = H5ADataSet(adata)
        else:
            geneWeights = None
            cell = H5ADataSet(adata)
            
        cell_loader = DataLoader(cell, batch_size=self.batch_size, num_workers=0, shuffle=True)
        
        pyro.clear_param_store()
        print('...')  

        if os.path.exists(os.path.join(target_dir, 'model_save', self.modelName + '_' + str(iteration-1) + '.pth')):
            vae = self.model
            # dis = self.discriminator
            if os.path.exists(os.path.join(target_dir, 'model_save', self.modelName + '_' + str(iteration) + '.pth')):
                print('load current iteration model....')
                # dis.load_state_dict(torch.load(os.path.join(target_dir, 'model_save/' + self.modelName + '_dis_' + str(iteration) + '.pth')))
                vae.load_state_dict(torch.load(os.path.join(target_dir, 'model_save/' + self.modelName + '_' + str(iteration) + '.pth')))
            else:
                print('load last iteration model.....')
            
                # dis.load_state_dict(torch.load(os.path.join(target_dir, 'model_save/' + self.modelName + '_dis_' + str(iteration-1) + '.pth')))
                vae.load_state_dict(torch.load(os.path.join(target_dir, 'model_save/' + self.modelName + '_' + str(iteration-1) + '.pth')))
        else:
            vae = self.model

        vae.to(self.device)
        if geneWeights is None and is_iterative:
            print('no geneWeight')
    
        gc.collect()

        train_elbo = []
        epoch_range = self.epoch_iter if is_iterative else self.epoch_initial
        if sp.isspmatrix(adata.X):
            adata.X = adata.X.toarray()
        for epoch in range(epoch_range):
            print(epoch)
            total_epoch_loss_train = self.train_model(adata, vae, cell_loader, adj, geneWeights=geneWeights if is_iterative else None, use_cuda=self.cuda)
            train_elbo.append(-total_epoch_loss_train)
            print("[epoch %03d]  average training loss: %.4f" % (epoch, total_epoch_loss_train))
            with open(os.path.join(target_dir, '%d/loss.txt' % (int(iteration))), "a+") as f:
                f.write("[epoch %03d]  average training loss: %.4f\n" % (epoch, total_epoch_loss_train))
                f.close()
        torch.save(vae.state_dict(), os.path.join(target_dir, 'model_save/' + self.modelName + '_' + str(iteration) + '.pth'))
        # torch.save(dis, os.path.join(target_dir, 'model_save/' + self.modelName + '_dis_' + str(iteration) + '.pth'))
        # torch.save(vae.state_dict(), os.path.join(target_dir, 'model_save/' + self.modelName + '_' + str(iteration) + '.pth'))
        # torch.save(dis.state_dict(), os.path.join(target_dir, 'model_save/' + self.modelName + '_dis_' + str(iteration) + '.pth'))
    


'''
The module contains the VAE model and the discriminator model.
'''
import numpy as np
import gc
import anndata
import pandas as pd
from torch.nn.modules.module import Module
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from pyro.primitives import param,deterministic
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
from scipy.stats import multivariate_normal
import pyro
import pyro.distributions as dist
from ..dynamic_graphs.distDistance import *
from scvi.module.base import PyroBaseModuleClass
import gc
import math
from sklearn.neighbors import kneighbors_graph

TTT = 0
from torch.distributions import constraints
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)
from torch.nn.functional import softplus

from pyro.distributions import TorchDistribution, LogNormal, Poisson, Gamma, Weibull, Chi2
from pyro.distributions.util import broadcast_shape
import torch
import torch.nn.functional as F
from torch.distributions import constraints
from torch.distributions.distribution import Distribution
from torch.distributions.utils import broadcast_all, probs_to_logits, lazy_property, logits_to_probs

from numbers import Number

import torch
from torch.distributions import constraints
from torch.distributions.exp_family import ExponentialFamily
from torch.distributions.utils import broadcast_all
class ZeroInflatedDistribution(TorchDistribution):
    """
    Generic Zero Inflated distribution.

    This can be used directly or can be used as a base class as e.g. for
    :class:`ZeroInflatedPoisson` and :class:`ZeroInflatedNegativeBinomial`.

    :param TorchDistribution base_dist: the base distribution.
    :param torch.Tensor gate: probability of extra zeros given via a Bernoulli distribution.
    :param torch.Tensor gate_logits: logits of extra zeros given via a Bernoulli distribution.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }

    def __init__(self, base_dist, *, gate=None, gate_logits=None, validate_args=None):
        if (gate is None) == (gate_logits is None):
            raise ValueError(
                "Either `gate` or `gate_logits` must be specified, but not both."
            )
        if gate is not None:
            batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
            self.gate = gate.expand(batch_shape)
        else:
            batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
            self.gate_logits = gate_logits.expand(batch_shape)
        if base_dist.event_shape:
            raise ValueError(
                "ZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(base_dist.event_shape)
            )

        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()
        
        super().__init__(batch_shape, event_shape, validate_args=False)

    

    @lazy_property
    def gate(self):
      
        return logits_to_probs(self.gate_logits, is_binary=True)

    @lazy_property
    def gate_logits(self):
        return probs_to_logits(self.gate, is_binary=True)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        
        if "gate" in self.__dict__:

            gate, value = broadcast_all(self.gate, value)
            temp_value = value.clone()
            temp_value[temp_value==0] = 1e-7
            log_prob = (-gate).log1p() + self.base_dist.log_prob(temp_value)
            log_prob = torch.where(value == 0, (gate).log(), log_prob)
        else:
            gate_logits, value = broadcast_all(self.gate_logits, value)
            temp_value = value.clone()
            temp_value[temp_value==0] = 1e-7
            temp_base_log_prob= self.base_dist.log_prob(temp_value)
            log_prob_minus_log_gate = -gate_logits + temp_base_log_prob#self.base_dist.log_prob(temp_value)
            log_gate = -softplus(-gate_logits)
            
            log_prob = log_prob_minus_log_gate + log_gate
            zero_log_prob = log_gate#-0.5#softplus(log_prob_minus_log_gate) + log_gate#+1
            log_prob = torch.where(value == 0, zero_log_prob, log_prob)
        return log_prob


    def sample(self, sample_shape=torch.Size()):
        
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
        return samples


    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean**2 + self.base_dist.variance
        ) - (self.mean) ** 2

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
        gate_logits = (
            self.gate_logits.expand(batch_shape)
            if "gate_logits" in self.__dict__
            else None
        )
        base_dist = self.base_dist.expand(batch_shape)
        ZeroInflatedDistribution.__init__(
            new, base_dist, gate=gate, gate_logits=gate_logits, validate_args=False
        )
        new._validate_args = self._validate_args
        return new
class myZeroInflatedGamma(ZeroInflatedDistribution):
    """
    A Zero Inflated Gamma distribution.

    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the half open interval [0, 1).
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        'loc': constraints.real, 'scale': constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
   
    support = constraints.positive

    def __init__(
        self,
        loc,
        scale=None,
        gate=None,
        gate_logits=None,
        validate_args=None
    ):
        base_dist = Gamma(concentration=loc,rate=scale,validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )

class myZeroInflatedLogNormal(ZeroInflatedDistribution):
    """
    A Zero Inflated Log Normal distribution.

    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the half open interval [0, 1).
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        'loc': constraints.real, 'scale': constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
   
    support = constraints.positive

    def __init__(
        self,
        loc,
        scale=None,
        gate=None,
        gate_logits=None,
        validate_args=None
    ):
        base_dist = LogNormal(loc=loc,scale=scale,validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )
        
    
class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        # print(input.shape,self.weight.shape)
        # support = self.softplus(self.fc(input.float()))
        # support = self.BN2(support)
        
        # # print(support.shape)
        # support = torch.mm(input.float(), self.weight.float())
        # output = torch.spmm(adj, support)
        # print(torch.sum(input))

        support = torch.spmm(adj, input)
        # print(torch.sum(support))
        
        # output = self.fc(support)
        # output = torch.mm(support, self.weight.float())
        return support, support 

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'

class GraphEncoder(nn.Module):
    '''
    The GCN encoder of VAE, which encodes the single-cell data into the latent space.

    parameters:
    in_dim: the input dimension of the encoder.
    z_dim: the latent dimension of the encoder.
    hidden_dim: the GCN dimension of the encoder.

    return:
    z_loc: the mean of the latent space.
    z_scale: the variance of the latent space.
    
    '''
    def __init__(self, in_dim, z_dim, hidden_dim):
        super().__init__()
        
        self.gc1 = GraphConvolution(in_dim, hidden_dim)
        # self.gc1 = GraphConvolution(in_dim, hidden_dim)
        # self.dropout = dropout
        #self.fc1 = nn.Linear(nhid, 512)
        #self.fc2 = nn.Linear(512, hidden_dim)
        #self.fc3 = nn.Linear(256, hidden_dim)
        self.fc21 = nn.Linear(hidden_dim, z_dim)
        self.fc22 = nn.Linear(hidden_dim, z_dim)
        #self.BN1 = nn.BatchNorm1d(hidden_dim)
        
        self.test = nn.Linear(in_dim, hidden_dim)
        #self.test1 = nn.Linear(hidden_dim, hidden_dim)
        self.softplus = nn.Softplus()

    def forward(self, x,adj,  start, end, test=False): 
        if test == False:
            y, x = self.gc1(x, adj)
            hidden = x.clone()
            # x = self.softplus(x)
            hidden1 = x
            # hidden=x[start:end,:] #temporarily remove to test full batch vs small batch
            # hidden = self.softplus(self.BN1(x)) #temporarily remove to test full batch vs small batch
            y=y[start:end,:]
            # hidden = self.softplus(self.test(x))
            # x = F.dropout(x, self.dropout, training=self.training)

            hidden1=x[start:end,:]
        else:
            y = x
            hidden1 = x.clone()
        
        # print(x)
        
        hidden1 = self.softplus(self.test(hidden1))
        #hidden1 = self.softplus(self.test1(hidden1))
        #hidden1=  self.softplus(self.test1(hidden1))
        # hidden = self.softplus(x)
        # hidden1 = self.softplus(self.fc1(x))
        # hidden1 = self.softplus(self.fc2(hidden1))
        #hidden1 = self.softplus(self.fc3(hidden1))
        # print(hidden)
        z_loc = self.fc21(hidden1)
        z_scale = torch.exp(torch.clamp(self.fc22(hidden1), -5, 5))
        #print(z_loc)
        return z_loc, z_scale
    
class Discriminator(nn.Module):
    '''
        The discriminator of UNAGI, which distinguishes the real data from the generated data.
    '''
    def __init__(self,input_dim,hidden_dim):
        super().__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, 1)
        #self.fc3 = nn.Linear(64, 1)
        self.ReLU6 = nn.Softplus()
        self.BN1 = nn.BatchNorm1d(hidden_dim)
        
        self.bce = nn.BCEWithLogitsLoss()
    def forward(self, x, y):
        ys = y
        xs = x
        xs = self.ReLU6(self.fc1(xs))#+1e-6

        #print(xs)
        #xs = self.ReLU6(self.fc2(xs))+1e-6
        xs = self.fc2(xs)

        #ys = torch.LongTensor(ys)
        return self.bce(xs,ys)
    
class Decoder(nn.Module):
    def __init__(self, in_dim, z_dim, hidden_dim):
        super().__init__()

        self.fc1 = nn.Linear(z_dim, hidden_dim)
        #self.fc2 = nn.Linear(512, hidden_dim)
        #self.test = nn.Linear(hidden_dim,hidden_dim)
        #self.fc3 = nn.Linear(512, hidden_dim)
        self.fc21 = nn.Linear(hidden_dim, in_dim)
        self.fc22 = nn.Linear(hidden_dim, in_dim)
        self.fc23 = nn.Linear(hidden_dim, in_dim)
        #self.fc24 = nn.Linear(hidden_dim, in_dim)
        self.softplus = nn.Softplus()
        self.ReLU6 = nn.ReLU6()
        self.ReLU = nn.ReLU()
        self.sigmoid = nn.Sigmoid()
    def forward(self, z):

        hidden =self.softplus(self.fc1(z))#+1e-7
        

        loc =  self.fc21(hidden) #mu

        scale = self.softplus(self.fc22(hidden))
        dropout = self.sigmoid(torch.clamp(self.fc23(hidden),-3,3))
        return loc, scale, dropout
    
class VAE(PyroBaseModuleClass):
    '''
    The VAE model of UNAGI, which contains the encoder and decoder of VAE.
    '''
    def __init__(self,  n_input, n_latent,n_graph,beta,distribution): 
        super().__init__()
        self.n_latent = n_latent
        self.n_input = n_input
        self.n_graph = n_graph
        self.beta = beta
        self.distribution = distribution
       # in the init, we create the parameters of our elementary stochastic computation unit.
       
        # First, we setup the parameters of the generative model
        self.decoder = Decoder(self.n_input,self.n_latent,self.n_graph)
        self.log_theta = torch.nn.Parameter(torch.randn(self.n_input))
        self.gate_logits = torch.nn.Parameter(torch.randn(self.n_input))
        self.discriminator = Discriminator(self.n_input, self.n_latent)
        # Second, we setup the parameters of the variational distribution
        self.encoder = GraphEncoder(self.n_input, self.n_latent,self.n_graph)
        # self.gc = GraphConvolution(n_input, nhid)
        # self.softplus = nn.Sigmoid()
        
    def model(self, x,adj,batch,  start, end):
        '''
        define the model p(x|z)p(z) using the generative network p(x|z) and prior p(z) = N(0,1). The distribution of x is chosen from the `dist` parameter in the `setup_training` function. The default is `ziln` (zero-inflated log normal).

        parameters:
        x: the single-cell data.
        adj: the cell graph.
        batch: the batch information of the single-cell data.
        start: the start index of the single-cell data.
        end: the end index of the single-cell data.

        '''
        # register PyTorch module `decoder` with Pyro p(z) p(x|z)
        pyro.module("decoder", self)
        with pyro.plate("data", x[start:end,:].shape[0]):

            # setup hyperparameters for prior p(z)
            z_loc = x.new_zeros(torch.Size((x[start:end,:].shape[0], self.n_latent)))
            z_scale = x.new_ones(torch.Size((x[start: end,:].shape[0], self.n_latent)))
            # z_p = Variable(torch.randn(64,128))
            # sample from prior (value will be sampled by guide when computing the ELBO)
            z = pyro.sample("latent", dist.Normal(z_loc, z_scale).to_event(1))
            # get the "normalized" mean of the negative binomial
            if self.distribution == 'ziln':
                dec_loc, dec_mu, dec_dropout  = self.decoder(z)
                scale = torch.exp(self.log_theta)
                loc = (dec_mu+1e-5).log()-scale.pow(2)/2
                x_dist = myZeroInflatedLogNormal(loc=loc, scale=scale, gate=dec_dropout)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)
            elif self.distribution == 'zinb':
                _, mu, dec_dropout  = self.decoder(z)
                theta = self.log_theta.exp()
                nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            
                x_dist = dist.ZeroInflatedNegativeBinomial(total_count=theta, logits=nb_logits,gate= dec_dropout, validate_args=False)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)
            elif self.distribution =='nb':
                _, mu, dec_dropout  = self.decoder(z)
                theta = self.log_theta.exp()
                nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            
                x_dist = dist.NegativeBinomial(total_count=theta, logits=nb_logits, validate_args=False)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)

            elif self.distribution =='zig':
                _, mu, dec_dropout = self.decode(z)
                scale = F.softplus(self.log_theta)#self.log_theta.exp()
                loc = mu*scale
                rx = myZeroInflatedGamma(loc=loc, scale=scale, gate=dec_dropout, validate_args=False)
                a = deterministic('recon',rx.mean)
            else:
                raise ValueError('distribution not supported')
            return rx
        
    def guide(self, x, adj, batch, start, end):
        '''
        define the guide (i.e. variational distribution) q(z|x)
        '''
        # register PyTorch module `encoder` with Pyro
        pyro.module("encoder", self)
        
        with pyro.plate("data", x[start:end,:].shape[0]):
            # use the encoder to get the parameters used to define q(z|x)
            #x_ = torch.log(1 + x)
            [qz_m,qz_v] = self.encoder(x, adj,  start, end)
          
            #qz_v = self.var_encoder(x_)
            # sample the latent code z
            rz=pyro.sample("latent", dist.Normal(qz_m, qz_v).to_event(1))
            return rz
        
    def getZ(self, x,adj,batch, start, end,test=True):
        '''
        Get the latent space of the single-cell data.

        parameters:
        x: the single-cell data.
        adj: the cell graph.
        batch: the batch information of the single-cell data.
        start: the start index of the single-cell data.
        end: the end index of the single-cell data.
        test: whether to use the test mode. Default is True.
        '''
        # x = torch.log(1+x)
        z_loc, z_scale = self.encoder(x,adj,  start, end,test=test)
        # sample in latent space
        # z = dist.Normal(z_loc, z_scale).sample()
        
        return z_loc+z_scale,z_loc, z_scale

    def generate1(self,x,adj,batch, start, end):
        '''
        Reconstruct the single-cell data from the latent space given inputs.

        parameters:
        x: the single-cell data.
        adj: the cell graph.
        batch: the batch information of the single-cell data.
        start: the start index of the single-cell data.
        end: the end index of the single-cell data.

        return:
        rx: the reconstructed single-cell data.
        '''
        # x = torch.log(1 + x)
        z_loc, z_scale = self.encoder(x,adj,batch,  start, end,test=True)
        z = dist.Normal(z_loc, z_scale).sample()
        if self.distribution == 'ziln':
            dec_loc, dec_mu, dec_dropout  = self.decoder(z_loc+z_scale)
            scale = torch.exp(self.log_theta)
            loc = (dec_mu+1e-5).log()-scale.pow(2)/2
            x_dist = myZeroInflatedLogNormal(loc=loc, scale=scale, gate=dec_dropout)
            rx=x_dist.mean #-> expectation
        elif self.distribution == 'zinb':
            _, mu, dec_dropout  = self.decoder(z_loc+z_scale)
            theta = self.log_theta.exp()
            nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            x_dist = dist.ZeroInflatedNegativeBinomial(total_count=theta, logits=nb_logits,gate= dec_dropout, validate_args=False)
            rx=x_dist.mean
        elif self.distribution =='zig':
            _, mu, dec_dropout = self.decoder(z_loc+z_scale)
            scale = F.softplus(self.log_theta)
            loc = mu*scale
            x_dist = myZeroInflatedGamma(loc=loc, scale=scale, gate=dec_dropout, validate_args=False)
            rx = x_dist.mean
        # rx=x_dist.sample() #-> sample
        elif self.distribution =='nb':
            _, mu, dec_dropout  = self.decoder(z_loc+z_scale)
            theta = self.log_theta.exp()
            nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            x_dist = dist.NegativeBinomial(total_count=theta, logits=nb_logits, validate_args=False)
            rx=x_dist.mean
        else:
            raise ValueError('distribution not supported')
        return  rx#, None, None,None, None


import scanpy as sc
import numpy as np
from sklearn.neighbors import kneighbors_graph
def get_neighbors(stagedata,num_cells,anchor_neighbors,max_neighbors,min_neighbors):
    '''
    get neighbor parameters for each satge

    parameters
    ----------
    stagedata: list 
        list of adata for each stage
    num_cells: list
        list of number of cells for each stage
    anchor_neighbors: int
        number of neighbors for anchor stage
    max_neighbors: int
        maximum number of neighbors for each stage
    min_neighbors: int
        minimum number of neighbors for each stage

    return  
    ----------
    neighbors: list
        list of number of neighbors for each stage
    anchor_index: int
        index of anchor stage in stagedata

    '''
    temp_num_cells = num_cells.copy()
    temp_num_cells.sort()
    median_index = int(len(temp_num_cells)/2)
    anchor = temp_num_cells[median_index]

    anchor_index = num_cells.index(anchor)
    temp_adata = stagedata[anchor_index]
    # distance = temp_adata.obsp['distances'].toarray()
    distance = temp_adata.obsp['distances'][temp_adata.obsp['distances'].nonzero()].reshape(len(temp_adata),-1)
    distance[:,::-1].sort()
    distance = distance[:,:max_neighbors]
    distance.sort()

    avg_anchor_distance = np.mean(distance[:,anchor_neighbors-1])
    neighbors = []
    for i in range(len(num_cells)):
        if i != anchor_index:
            temp_adata = stagedata[i]
            #distance = temp_adata.obsp['distances'].toarray()
            distance = temp_adata.obsp['distances'][temp_adata.obsp['distances'].nonzero()].reshape(len(temp_adata),-1)
            distance[:,::-1].sort()
            distance = distance[:,:max_neighbors]
            distance.sort()
            search = []
            for neighbor in range(max_neighbors):
                search.append(abs(np.mean(distance[:,neighbor])-avg_anchor_distance))
            #find the index of minimum value in search
            min_index = search.index(min(search))+1
            if min_index<min_neighbors:
                min_index=min_neighbors

            
            neighbors.append(min_index)
        else:
            neighbors.append(anchor_neighbors)
    return neighbors, anchor_index

def get_mean_median_cell_population(adata):
    '''
    get the mean and median number of cells in each cluster

    parameters
    ----------
    adata: anndata
        anndata of the stage

    return
    ----------
    mean: float
        mean number of cells of each cluster
    median: float
        median number of cells of each cluster
    '''
    num_cells = []
    for i in list(adata.obs['leiden'].unique()):
        temp = adata.obs[adata.obs['leiden'] == i].index.tolist()
        temp = adata[temp]
        num_cells.append(len(temp))
    return np.mean(num_cells)/len(adata), np.median(num_cells)/len(adata)
def auto_resolution(stagedata, anchor_index,neighbors, min_res, max_res):
    '''
    get the optimal resolution for each stage

    parameters
    ----------
    stagedata: list
        list of adata for each stage
    anchor_index: int
        index of anchor stage in stagedata
    neighbors: list
        list of number of neighbors for each stage
    min_res: float
        minimum resolution for leiden clustering
    max_res: float
        maximum resolution for leiden clustering

    Return:

    --------------
    
    out_res: list
        list of optimal resolution for each stage
    all_means: list
        list of mean number of cells in each cluster for each stage
    '''
    anchor_adata = stagedata[anchor_index]
    anchor_adata.obsp['connectivities'] = kneighbors_graph(anchor_adata.obsm['z'], neighbors[anchor_index], mode='connectivity', include_self=True,n_jobs=20)
    anchor_adata.obsp['distances'] = kneighbors_graph(anchor_adata.obsm['z'], neighbors[anchor_index], mode='distance', include_self=True,n_jobs=20)
    sc.tl.leiden(anchor_adata, resolution = min_res)
    all_means = []
    anchor_mean, anchor_median = get_mean_median_cell_population(anchor_adata)
    out_res = []
    for i in range(len(stagedata)):
        differences = []
        temp_all_means = []
        if i != anchor_index:
            temp_adata = stagedata[i]
            temp_adata.obsp['connectivities'] = kneighbors_graph(temp_adata.obsm['z'], neighbors[i], mode='connectivity', include_self=True,n_jobs=20)
            temp_adata.obsp['distances'] = kneighbors_graph(temp_adata.obsm['z'], neighbors[i], mode='distance', include_self=True,n_jobs=20)
            for j in np.arange(min_res,max_res+0.1,0.1):
                sc.tl.leiden(temp_adata, resolution = j)
                temp_mean, temp_median = get_mean_median_cell_population(temp_adata)
                temp_all_means.append(temp_mean)
                differences.append(abs(temp_mean-anchor_mean))
            min_index = differences.index(min(differences))
            out_res.append(min_index*0.1+min_res)
            all_means.append(temp_all_means[min_index])
        else:
            out_res.append(1)
            all_means.append(anchor_mean)
    return out_res, all_means


import subprocess
from ..utils.gcn_utils import  setup_graph
from ..utils.CPO_utils import get_neighbors, auto_resolution
from ..utils.attribute_utils import saveRep,get_all_adj_adata,mergeAdata,updateAttributes,get_data_file_path
from ..dynamic_graphs.buildGraph import getandUpadateEdges
import gc
import scanpy as sc
import numpy as np
import os
from ..dynamic_regulatory_networks.processIDREM import getClusterPaths, getClusterIdrem, runIdrem
from ..dynamic_regulatory_networks.processTFs import getTFs, getTargetGenes, matchTFandTGWithFoldChange, updataGeneTablesWithDecay
from sklearn.neighbors import kneighbors_graph
from scipy.sparse import csr_matrix
class UNAGI_runner:
    '''
    The UNAGI_runner class is used to set up the hyperparameters to run iDREM, find clustering optimal parameters and run the UNAGI model . It takes the following 
    
    parameters
    ------------
    data_path: the path to the data
    total_stage: the total number of time-series stages
    iteration: the total iteration to run the UNAGI model
    trainer: the trainer class to train the UNAGI model
    idrem_dir: the directory of the idrem software
    '''
    def __init__(self,data_path,total_stage,iteration,trainer,idrem_dir):
        self.data_path = data_path
        self.total_stage = total_stage
        self.iteration = iteration
        self.trainer = trainer
        self.resolutions = None
        self.idrem_dir = idrem_dir
        self.neighbor_parameters = None
        self.setup_CPO = False
        self.species = None
        self.setup_IDREM = False
    def load_stage_data(self):
        '''
        Load the stage data from the data_path. The stage data will be stored in the adata_stages list. The all_in_one adata will be used for the UNAGI model training.
        '''
        stageadata = []
        for i in range(self.total_stage):
            if self.iteration != 0:
                adata = sc.read(os.path.join(self.data_path, '%d/stagedata/%d.h5ad'%(self.iteration-1,i)))
            elif self.iteration == 0:
                adata = sc.read(os.path.join(self.data_path, '%d.h5ad'%(i)))
            if 'leiden_colors' in adata.uns:
                del adata.uns['leiden_colors']
            stageadata.append(adata)
        self.all_in_one = get_all_adj_adata(stageadata)
        self.adata_stages = stageadata
        
    def annotate_stage_data(self, adata,stage):
        '''
        Retreive the latent representations of given single cell data. Performing clusterings, generating the UMAPs, annotating the cell types and adding the top genes and cell types attributes.

        Parameters
        ------------
        adata: anndata
            the single cell data.
        stage: int
            the stage of the single cell data.

        return
        ------------
        adata: anndata
            the annotated single cell data.
        averageValue: list
            the average value of each cluster.
        reps: list
            the latent representations of the single cell data.
        '''
        z_locs, z_scales, cell_embeddings = self.trainer.get_latent_representation(adata,self.iteration,self.data_path)
        adata.obsm['z'] = cell_embeddings
        if self.neighbor_parameters is None:
            sc.pp.neighbors(adata, use_rep="z",n_neighbors=50,method='umap')
            return adata
        if 'connectivities' in adata.obsp.keys():
            del adata.obsp['connectivities']
            del adata.obsp['distances']
        z_adj = kneighbors_graph(adata.obsm['z'], self.neighbor_parameters[stage], mode='connectivity', include_self=True,n_jobs=20)
        adata.obsp['connectivities'] = z_adj
        adata.obsp['distances'] = kneighbors_graph(adata.obsm['z'], self.neighbor_parameters[stage], mode='distance', include_self=True,n_jobs=20)
        sc.tl.leiden(adata,resolution = self.resolutions[stage])
        sc.tl.paga(adata)
        sc.pl.paga(adata,show=False)
        # sc.tl.umap(adata, min_dist=0.25/self.resolution_coefficient*len(adata),init_pos='paga')
        sc.tl.umap(adata, min_dist=0.05,init_pos='paga')
        rep=[z_locs, z_scales]
        adata.obs['ident'] = 'None'
        adata.obs['leiden'] = adata.obs['leiden'].astype(str)
        sc.pl.umap(adata,color='leiden',show=False)
        adata,averageValue,reps = updateAttributes(adata,rep)#add top genes, cell types attributes and get average cluster value
  
        adata.obs['leiden']=adata.obs['leiden'].astype(str)
        if self.iteration == 0:
            allzeros = np.zeros_like(adata.X)
            allzeros = csr_matrix(allzeros)
            adata.layers['geneWeight'] = allzeros
        sc.pl.umap(adata,color='ident',show=False)
        adata.write(os.path.join(self.data_path,str(self.iteration)+'/stagedata/%d.h5ad'%stage),compression='gzip',compression_opts=9)
        return adata,averageValue,reps
    
    def set_up_CPO(self, anchor_neighbors, max_neighbors, min_neighbors, resolution_min, resolution_max):
        '''
        Set up the parameters for finding the clustering optimal parameters.

        Parameters
        ------------
        anchor_neighbors: int
            the number of neighbors for the anchor cells.
        max_neighbors: int
            the maximum number of neighbors for the single cell data.
        min_neighbors: int
            the minimum number of neighbors for the single cell data.
        resolution_min: float
            the minimum resolution for the single cell data.
        resolution_max: float   
            the maximum resolution for the single cell data.
        '''
        self.setup_CPO = True
        self.anchor_neighbors = anchor_neighbors
        self.max_neighbors = max_neighbors
        self.min_neighbors = min_neighbors
        self.resolution_min = resolution_min
        self.resolution_max = resolution_max

    def run_CPO(self):
        '''
        Find the clustering optimal parameters for the single cell data.
        '''
        max_adata_cells = 0
        num_cells = []
        for each in self.adata_stages:
            num_cells.append(each.shape[0])
            if len(each) > max_adata_cells:
                max_adata_cells = len(each)
        self.resolution_coefficient = max_adata_cells
        for i in range(0,len(self.adata_stages)):
            self.adata_stages[i] = self.annotate_stage_data(self.adata_stages[i], i)
        if not self.setup_CPO:
            print('CPO parameters are not set up, using default parameters')
            print('anchor_neighbors: 15, max_neighbors: 35, min_neighbors: 10, resolution_min: 0.8, resolution_max: 1.5')
            self.neighbor_parameters, anchor_index = get_neighbors(self.adata_stages, num_cells,anchor_neighbors=15,max_neighbors=35,min_neighbors=10)
            self.resolutions,_ = auto_resolution(self.adata_stages, anchor_index, self.neighbor_parameters, 0.8, 1.5)
        else:
            self.neighbor_parameters, anchor_index = get_neighbors(self.adata_stages, num_cells,anchor_neighbors=self.anchor_neighbors,max_neighbors=self.max_neighbors,min_neighbors=self.min_neighbors)
            self.resolutions,_ = auto_resolution(self.adata_stages, anchor_index, self.neighbor_parameters, self.resolution_min, self.resolution_max)
    
    def update_cell_attributes(self):
        '''
        Update and save the cell attributes including the top genes, cell types and latent representations.
        '''
        self.averageValues = []
        reps = []
        for i in range(0,len(self.adata_stages)):
            adata = self.adata_stages[i]
            # print(adata.X.shape)
            adata.uns['topGene']={}
            adata.uns['clusterType']={}
            adata.uns['rep']={}
            adata,averageValue,rep = self.annotate_stage_data(adata,i)
            gc.collect()
            reps.append(rep)
            averageValue = np.array(averageValue)
            self.averageValues.append(averageValue)
        self.averageValues = np.array(self.averageValues,dtype=object)
        np.save(os.path.join(self.data_path, '%d/averageValues.npy'%self.iteration),self.averageValues)
        saveRep(reps,self.data_path,self.iteration)
    def build_temporal_dynamics_graph(self):
        '''
        Build the temporal dynamics graph.
        '''
        self.edges = getandUpadateEdges(self.total_stage,self.data_path,self.iteration)
    def set_up_IDREM(self,Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood, Minimum_Standard_Deviation):
        '''
        Set up the parameters for running the iDREM software.

        Parameters
        ------------
        Minimum_Absolute_Log_Ratio_Expression: float
            the minimum absolute log ratio expression.
        Convergence_Likelihood: float
            the convergence likelihood.
        Minimum_Standard_Deviation: float
            the minimum standard deviation.
        '''
        self.setup_IDREM = True
        self.Minimum_Absolute_Log_Ratio_Expression = Minimum_Absolute_Log_Ratio_Expression
        self.Convergence_Likelihood = Convergence_Likelihood
        self.Minimum_Standard_Deviation = Minimum_Standard_Deviation
    def set_up_species(self, species):
        '''
        Set up the species for running the iDREM software.

        Parameters
        ------------
        species: str
            the species of the single cell data.
        '''

        print('Species: Running on %s data'%species)
        self.species = species
    def run_IDREM(self):
        '''
        Run the iDREM software.
        '''
        averageValues = np.load(os.path.join(self.data_path, '%d/averageValues.npy'%self.iteration),allow_pickle=True)
        paths = getClusterPaths(self.edges,self.total_stage)
        idrem= getClusterIdrem(paths,averageValues,self.total_stage)
        paths = [each for each in paths.values() if len(each) == self.total_stage]
        # print(paths) 
        idrem = np.array(idrem)
        self.genenames =np.array(list(self.adata_stages[0].var.index.values))
        if not self.setup_IDREM:
            print('IDREM parameters are not set up, using default parameters')
            print('Minimum_Absolute_Log_Ratio_Expression: 0.5, Convergence_Likelihood: 0.001, Minimum_Standard_Deviation: 0.5')
            if self.species is None:
                print('Human species is used as default')
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir)
            else:
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,species=self.species)
        else:
            if self.species is None:
                print('Human species is used as default')
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,Minimum_Absolute_Log_Ratio_Expression=self.Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood=self.Convergence_Likelihood, Minimum_Standard_Deviation=self.Minimum_Standard_Deviation)
            else:
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,species=self.species,Minimum_Absolute_Log_Ratio_Expression=self.Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood=self.Convergence_Likelihood, Minimum_Standard_Deviation=self.Minimum_Standard_Deviation)
            
    def update_gene_weights_table(self,topN=100):
        '''
        Update the gene weights table.

        Parameters
        ------------
        topN: int
            the number of top genes to be selected.
        
        '''
        TFs = getTFs(os.path.join(self.data_path,str(self.iteration)+'/'+'idremResults'+'/'),total_stage=self.total_stage)
        scope = getTargetGenes(os.path.join(self.data_path,str(self.iteration)+'/'+'idremResults'+'/'),topN)
        p = matchTFandTGWithFoldChange(TFs,scope,self.averageValues,get_data_file_path('human_encode.txt'),self.genenames,self.total_stage)
        #np.save('../data/mes/'+str(iteration)+'/tfinfo.npy',np.array(p))
        updateLoss = updataGeneTablesWithDecay(self.data_path,str(self.iteration),p,self.total_stage)
    def build_iteration_dataset(self):
        '''
        Build the iteration dataset.
        '''
        mergeAdata(os.path.join(self.data_path,str(self.iteration)),total_stages=self.total_stage)
       
    def run(self):
        '''
        Run the UNAGI pipeline.
        '''
        self.load_stage_data()
        if self.iteration == 0:
            is_iterative = False
        else:
            is_iterative = True
        self.trainer.train(self.all_in_one,self.iteration,target_dir=self.data_path,is_iterative=is_iterative)
        self.run_CPO()
        self.update_cell_attributes()
        self.build_temporal_dynamics_graph()
        self.run_IDREM()
        self.update_gene_weights_table()
        self.build_iteration_dataset()





import multiprocessing
import numpy as np
import gc
import anndata
import time
import pandas as pd
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
from scipy.stats import multivariate_normal
import gc

def getSimilarity(adatai,adataj, i, j):
    '''
    calculate the differential gene similarity between two clusters. Differentail gene similarity = (1-Jaccard index) * gene ranking difference factor
    
    parameters
    -------------------
    adatai: anndata
        The data of stage i
    adataj: anndata 
        The data of stage j
    i: int
        the cluster id in stage i
    j: int
        the cluster id in stage j

    return
    -------------------
    distance: np.array
        The distance of top 100 differential gene between two clusters

    '''
  
    if isinstance(adatai[i], list):
        
        iDE = adatai[i]
    else:
        iDE = adatai[i].tolist()
    if isinstance(adataj[j], list):
        
        jDE = adataj[j]
    else:
        jDE = adataj[j].tolist()
  

    intersection = 0
    orderdistance = 0
    for i, each in enumerate(iDE):
        if each in jDE:
            
            intersection+=1
            jindex = jDE.index(each)
        
            orderdistance += abs(jindex-i)
        else:
            orderdistance+=100
    union = len(iDE)+len(jDE) - intersection
    JaccardIndex = intersection/union
   
    orderdistance = orderdistance/(100*len(jDE))
    # print(orderdistance)
    DEG_distance = (1-JaccardIndex)*orderdistance
    return DEG_distance

def mcSampling(mus, sigmas):
    '''
    monte carlo sampling strategy to sample data points from a gaussian distribution in each cell,
    for example the hidden space is 10, then sample 100 data point from input cell. If the number of cell is 200, then
    the sampled matrix will be [10, 200*100]
    
    parameters
    -------------------
    mus: list
        mu vectors of input cells (shape: [number of cell, number of hidden nodes])

    sigmas: list
        sigma vectors of input cells (shape: [number of cell, number of hidden nodes])

    
    return
    -------------------
    samplegaussian: list
        A list of sampled data-points from fitted gaussian distributions.
    '''
    samplegaussian= [[] for _ in range(64)] #hyperparameter needed to be changed by users
    hiddensize=64#len(mus[0])
    mus = np.vstack(mus)
    sigmas=np.vstack(sigmas)

    for i in range(hiddensize):
        
        mean = mus[:,i]
        mean = mean.repeat(100)
        std = sigmas[:,i]
        std = std.repeat(100)
        normal = Normal(torch.zeros(size=mean.shape)+torch.tensor(mean),torch.zeros(size=std.shape)+torch.tensor(std)).sample()
        samplegaussian[i].append(normal.numpy())

    return samplegaussian

class GaussianRepThread(threading.Thread):
    '''
    The class to fit gaussian distributions for each hidden node.

    parameters
    -------------------
    output: list
        A list of threads
    data: list
        A list of samples of the gaussian distribution of each hidden node
    i: int
        the index of the thread
    '''
    def __init__(self, output, data,i):
        threading.Thread.__init__(self)
        self.i = i
        self.data = data
        self.output = output
    def run(self):
        #time.sleep(10)
        self.output[self.i] = norm.fit(self.data)
def fitClusterGaussianRepresentation(data):
    '''
    Fitting gaussian distributions for each hidden node

    parameters
    -------------------
    data: list
        A list of samples of the gaussian distribution of each hidden node
    
    return
    -------------------
    out: list
        A list of mu and sigma of each hidden node
    '''
    threads = []
    out = [[] for _ in range(len(data))]
    for i in range(len(data)):
        threads.append(GaussianRepThread(out,data[i],i))
    for each in threads:
        each.start()
    for each in threads:
        each.join()
    return out
def getClusterRepresentation(mus, sigmas):
    '''
    MC strategy to sample gaussian data points. Use sampled data points to fit gaussian distributions
    
    parameters
    -------------------
    mus: np.array
        mu vectors of input cells (shape: [number of cell, number of hidden nodes])
    
    sigmas: np.array
        sigma vectors of input cells (shape: [number of cell, number of hidden nodes])
   
    return
    -------------------
    fitted_gaussian_distributions: list
        A list of fitted gaussian distributions
    '''
    T1 = time.time()
    sampling = mcSampling(mus,sigmas)
    T2 = time.time()
    #print('time is %s ' %((T2-T1)))
    fitted_gaussian_distributions = fitClusterGaussianRepresentation(sampling)
    return fitted_gaussian_distributions

def normalizeDistance(distance):
    '''
    Normalize the kl divergence distance and top differential gene distances. (Use the min-max normalization method.)
    
    parameters
    -------------------
    distance: list
        A list of distance metrics. ([gaussian kl divergence, top differential gene difference])
    
    return
    -------------------

    normalizedDistance: np.array
        Sum of normalized kl divergence and top differential gene difference distance
    '''
    

    distance = np.array(distance)
    
    gaussiankld = np.array(distance[:,0])

    de = np.array(distance[:,1])
    if gaussiankld.shape[0] != 1:
        #min max normalization
        de_minmax = (de - de.min()) / (de.max() - de.min())
        ggkld = gaussiankld
        #min max normalization
        ggkld_minmax = (ggkld - ggkld.min()) / (ggkld.max() - ggkld.min())
    else:
        ggkld_minmax= gaussiankld
        de_minmax = de
    
    normalizedDistance = ggkld_minmax+de_minmax
    return normalizedDistance
def calculateKL(cluster1gaussian, cluster2gaussian):
    '''
    calculate KL divergence of multivariate gaussian distributions between two clusters.
    
    parameters
    ------------------- 

    cluster1gaussian: list
        A list of [mean, std] of gaussian distribution of cluster 1

    cluster2gaussian: list
        A list of [mean, std] of gaussian distribution of cluster 2

    
    return
    -------------------

    kl: np.array
        kl divergence of multivariate gaussian distributions between two clusters.
    '''
    cluster1 = np.array(cluster1gaussian,dtype=np.float64).reshape(-1,2)
    cluster2 = np.array(cluster2gaussian,dtype=np.float64).reshape(-1,2)
    
    std1 = cluster1[:,1]
    covariance1 = torch.tensor(np.diag(std1**2))

    mean1 = cluster1[:,0]

    p = MultivariateNormal(torch.tensor(mean1), covariance1)
    mean2 = cluster2[:,0]
    std2 = cluster2[:,1]
    covariance2 = torch.tensor(np.diag(std2**2))
    q = MultivariateNormal(torch.tensor(mean2), covariance2)
    kl = torch.distributions.kl.kl_divergence(p, q)
    
    kl = kl.detach().numpy()

    return kl








'''
this script is used to shuffle the gene expression for each stage data and use the shuffled data to build the random background to calculate the p-val for dynamic markers
'''
import scanpy as sc
import numpy as np
from ..dynamic_regulatory_networks.processIDREM import getClusterPaths,getClusterIdrem
def averageNode(nodes,state):
    '''
    calculate the average gene expression of sibling nodes
    args: 
    nodes: number of sibling nodes
    state: the gene expression of each cluster in a certain stage
    
    return: 
    out: the average gene expression of sibling nodes
    '''
    out = 0
    for each in nodes:
        out+=state[each]
    return out/len(nodes)
def get_progressionmarker_background(times, adata,total_stage):
    '''
    sampling and simulate the random background for dynamic markers.
    args:
    times: the number of times to sample the random background
    adata: the single-cell data
    total_stage: the total number of time stages

    return:
    results: the simulated random background
    '''
    results = {}
    edges = adata.uns['edges']
    paths = getClusterPaths(edges,total_stage)
    
    stage_data =[]
    for i in range(len(adata.obs['stage'].unique())):
        adata.obs['stage'] = adata.obs['stage'].astype(str)
        temp = adata[adata.obs[adata.obs['stage']==str(i)].index.tolist()]
        stage_data.append(temp)
    for time in range(times):
        print('shuffled times:',time)
        # results.append([])
        #shuffle the gene expression for each stage data and each gene
        avg_expression_val = []
        for i in range(len(stage_data)):
            avg_expression_val.append([])
            tempX = stage_data[i].X.toarray()
          
            # for j in range(stage_data[i].shape[1]):
            #     np.random.shuffle(tempX[:,j])
            np.random.shuffle(tempX)
            stage_data[i].layers['shuffle_X'] = tempX
            for cluster in range(len(stage_data[i].obs['leiden'].unique())):
                avg_expression_val[i].append(np.mean(stage_data[i][stage_data[i].obs[stage_data[i].obs['leiden']==str(cluster)].index.tolist()].layers['shuffle_X'],axis=0))
            del stage_data[i].layers['shuffle_X']

        idrem = np.array(getClusterIdrem(paths,avg_expression_val,total_stage))
        path = [each for each in paths.values() if len(each) == total_stage]
        for i, each in enumerate(path):
            track_name = ''
            for j, stage in enumerate(each):
                if j > 0:
                    track_name += '-'
                for k, cluster in enumerate(stage):
                    if k > 0:
                        track_name += 'n'
                    track_name += str(cluster)
            if track_name not in results.keys():
                results[track_name] = []
            results[track_name].append(idrem[i][:,-1] - idrem[i][:,0])
    return results
# out = permutation(permutation_time,adata)
# out = np.array(out).reshape(-1,adata.X.shape[1])
# print(out.shape)
# np.save('progressionMarker_background.npy',out)